from functools import wraps

import requests
from urecord import Record
from urlobject import URLObject, encode_component

from conneg.accept import make_accept_header
from conneg.default import DEFAULT_REGISTER
from http import requests_http_backend
from utils.caseless_dict import CaselessDictionary

__all__ = ['API']


def replaces_url(meth):
    """Wrap a method on :class:`API` to return a new API with a new URL."""

    @wraps(meth)
    def wrapper(self, *args, **kwargs):
        url = meth(self, *args, **kwargs)
        # If the method is already reflexive, return the result directly.
        if isinstance(url, type(self)):
            return url
        return self._with_url(url)
    return wrapper


# We're using Record here for efficiency and immutability.
class API(Record('url', 'http')):

    """
    Represents a remote HTTP-accessible resource (and operations thereon).

    Example usage::

        >>> api = API(u'https://graph.facebook.com/')
        >>> api
        <intessa.API(u'https://graph.facebook.com/')>

    Build paths using attribute access or subscription syntax::

        >>> api[u'19292868552']
        <intessa.API(u'https://graph.facebook.com/19292868552')>
        >>> api.cocacola
        <intessa.API(u'https://graph.facebook.com/cocacola')>
        >>> api[u'btaylor'][u'picture']
        <intessa.API(u'https://graph.facebook.com/btaylor/picture')>

    Perform requests and receive :class:`responses <intessa.response.Response>`
    by calling the object::

        >>> api[u'19292868552']()  # doctest: +SKIP
        <intessa.Response[200, text/javascript]>

    """

    def __new__(cls, url, http=requests_http_backend):
        if not isinstance(url, URLObject):
            url = URLObject.parse(url)
        return super(API, cls).__new__(cls, url, http)

    def __repr__(self):
        return '<intessa.API(%r)>' % (unicode(self),)

    def __unicode__(self):
        return unicode(self.url)

    def __eq__(self, other):
        return unicode(self) == other

    def __hash__(self):
        return hash(unicode(self.url))

    @replaces_url
    def __add__(self, other):

        """
        Simple string concatenation is reflexive.

            >>> API(u'http://example.com/') + u'foo'
            <intessa.API(u'http://example.com/foo')>
        """

        return unicode(self) + other

    @replaces_url
    def __and__(self, params):

        """
        Add parameters to this URL.

        Accepts pairs or dictionaries:

            >>> API(u'http://example.com/') & ('a', 'b')
            <intessa.API(u'http://example.com/?a=b')>
            >>> API(u'http://example.com/') & {'a': 'b'}
            <intessa.API(u'http://example.com/?a=b')>

        Multiple additions of the same parameter will keep all of them, in the
        order they were specified:

            >>> API(u'http://example.com/') & ('a', 'b') & ('a', 'c')
            <intessa.API(u'http://example.com/?a=b&a=c')>
        """

        return self.url & params

    @replaces_url
    def __or__(self, params):

        """
        Add parameters to this URL, replacing existing parameters.

        Accepts pairs or dictionaries:

            >>> API(u'http://example.com/') | ('a', 'b')
            <intessa.API(u'http://example.com/?a=b')>
            >>> API(u'http://example.com/') | {'a': 'b'}
            <intessa.API(u'http://example.com/?a=b')>

        Multiple additions of the same parameter will replace old values:

            >>> API(u'http://example.com/') | ('a', 'b') | ('a', 'c')
            <intessa.API(u'http://example.com/?a=c')>
        """

        return self.url | params

    @replaces_url
    def __getattr__(self, attr):

        """
        Dynamic attribute access adds path components to the current URL.

            >>> API(u'http://example.com/').users
            <intessa.API(u'http://example.com/users')>
        """

        return self[attr]

    @replaces_url
    def __getitem__(self, item):

        """
        Subscription syntax is overridden to extend or modify the current URL.

        Example usage::

            >>> api = API(u'http://example.com/')

        In the simplest case, add a path component::

            >>> api['foo']
            <intessa.API(u'http://example.com/foo')>
            >>> api['foo']['bar']
            <intessa.API(u'http://example.com/foo/bar')>

        Replace the whole path by prefixing with a '/'::

            >>> api['foo']['bar']['/baz']
            <intessa.API(u'http://example.com/baz')>

        Add or modify a fragment identifier::

            >>> api['#foo']
            <intessa.API(u'http://example.com/#foo')>
            >>> api['#foo']['#bar']
            <intessa.API(u'http://example.com/#bar')>

        Add or modify a file extension::

            >>> api['foo']['.json']
            <intessa.API(u'http://example.com/foo.json')>
            >>> api['foo']['.json']['.xml']
            <intessa.API(u'http://example.com/foo.xml')>

        Adding a file extension to a non-leaf node will add an 'index'
        component::

            >>> api['.json']
            <intessa.API(u'http://example.com/index.json')>

        Any necessary escaping will be taken care of::

            >>> api['foo bar baz']
            <intessa.API(u'http://example.com/foo%20bar%20baz')>
            >>> api['#foo bar#baz']
            <intessa.API(u'http://example.com/#foo%20bar%23baz')>
            >>> api['foo']['.js o%n']
            <intessa.API(u'http://example.com/foo.js%20o%25n')>
        """

        if not isinstance(item, basestring):
            item = unicode(item)

        # Replace whole path with a new one.
        if item.startswith('/'):
            return self.url.with_path(item[1:])

        # Add/replace fragment identifier.
        if item.startswith('#'):
            return self.url.with_fragment(item[1:])

        # Add/replace file extension. Add an 'index' path component if current
        # path is not a leaf node.
        if item.startswith('.'):
            if self.url.endswith('/'):
                return self['index'][item]
            return self._with_file_ext(item[1:])

        # Standard case of adding a path component.
        return self.url / item

    def _with_url(self, url):
        """Replace the URL on this API with another."""

        if not isinstance(url, URLObject):
            url = URLObject.parse(url)
        return self._replace(url=url)

    @replaces_url
    def _with_file_ext(self, file_ext):

        """
        Add/replace a file extension on this URL.

        Example::

            >>> API(u'http://example.com/foo')._with_file_ext('.json')
            <intessa.API(u'http://example.com/foo.json')>

        The leading period is optional::

            >>> API(u'http://example.com/foo')._with_file_ext('json')
            <intessa.API(u'http://example.com/foo.json')>

        Existing file extensions will be replaced::

            >>> API(u'http://example.com/foo.json')._with_file_ext('xml')
            <intessa.API(u'http://example.com/foo.xml')>

        This method only works on leaf nodes::

            >>> API(u'http://example.com/foo/')._with_file_ext('json')
            Traceback (most recent call last):
                ...
            ValueError: Cannot add a file extension to directories.

        """

        if self.url.endswith('/'):
            raise ValueError("Cannot add a file extension to directories.")

        # 'https://example.com/foo/bar/baz.xyz' => 'baz.xyz'
        leaf_node = self.url.path_list()[-1]
        # 'baz.xyz' => 'baz'
        base = leaf_node.rsplit('.', 1)[0]

        if not file_ext.startswith('.'):
            file_ext = '.' + file_ext

        return self.url.parent() / (base + file_ext)

    def __call__(self, *args, **kwargs):

        """
        Perform an HTTP request and return a :class:`~intessa.response.Response`.

        :param method:
            The HTTP method to use (defaults to ``GET``).

        :param data:
            A Python object representing the body of the request.

            There are several valid values this parameter can take:

            *   A bytestring, with a ``Content-Type`` header (given in
                ``headers``).
            *   A file-like object, with ``Content-Type`` and
                ``Content-Length`` headers.
            *   Any other Python object, with the ``type`` parameter provided.

            In the first two instances, it will be sent as-is. In the second,
            it will first be encoded to a bytestring, using the codec for the
            specified type.

        :param type:
            The Internet media type with which the ``data`` argument should be
            encoded. This will be used as a lookup type against the
            ``codec_register``.

        :param codec_register:
            A :class:`~intessa.conneg.codec_base.CodecRegister` to use for
            encoding the request body and decoding the response. Defaults to
            :data:`~intessa.conneg.default.DEFAULT_REGISTER`.

        :param accept:
            A specification of acceptable response types, which will be turned
            into an HTTP ``Accept`` header.

            This argument can be a string representing a media type (see:
            :class:`~intessa.conneg.content_type.ContentType`), a list of such
            strings, or a list of ``(media_type, quality)`` pairs, where
            ``quality`` is a floating-point number. See the documentation for
            :func:`~intessa.conneg.accept.make_accept_header` for examples of
            how this argument is turned into an ``Accept`` header.

        :param headers:
            A dictionary of raw HTTP headers to send.

        Note that if a non-successful response is returned, an instance of
        :class:`intessa.Error` will be *raised*.
        """

        method = kwargs.pop('method', 'GET')
        headers = CaselessDictionary(kwargs.pop('headers', {}))
        codec_register = kwargs.pop('codec_register', DEFAULT_REGISTER)

        if 'data' in kwargs:
            if method == 'GET':
                method = 'POST'
            data = kwargs.pop('data')
            type = kwargs.pop('type', None)
            headers, kwargs['data'] = encode_request(
                headers, data, type, codec_register=codec_register)

        accept = kwargs.pop('accept', None)
        if accept is not None:
            headers['Accept'] = make_accept_header(accept)

        kwargs['headers'] = headers
        return self.http(self, method, self.url, *args, **kwargs)

    def _post(self, *args, **kwargs):
        """Shim for issuing POST requests."""

        kwargs.setdefault('method', 'POST')
        return self(*args, **kwargs)

    def _put(self, *args, **kwargs):
        """Shim for issuing PUT requests."""

        kwargs.setdefault('method', 'PUT')
        return self(*args, **kwargs)

    def _delete(self, *args, **kwargs):
        """Shim for issuing DELETE requests."""

        kwargs.setdefault('method', 'DELETE')
        return self(*args, **kwargs)


def encode_request(headers, data, type=None, codec_register=DEFAULT_REGISTER):

    """
    Encode an object and a type into a bytestring/stream and HTTP headers.

    :param headers:
        The HTTP headers for the request. A new dictionary of headers will be
        returned which will contain a ``Content-Type``.

    :param data:
        A Python object representing the body of the request.

        There are several valid values this parameter can take:

        *   A bytestring (with a ``Content-Type`` header given in ``headers``).
        *   A file-like object (with ``Content-Type`` and ``Content-Length``
            headers).
        *   Any other Python object, with the ``type`` parameter provided.

        In the first two instances, it will be sent as-is. In the second,
        it will first be encoded to a bytestring, using the codec for the
        specified type.

    :param type:
        The Internet media type with which ``data`` should be encoded. This
        will be used as a lookup type against the ``codec_register``. If this
        argument is omitted, there *must* be a specified ``Content-Type`` in
        ``headers``, otherwise a ``TypeError`` will be raised.

    :param codec_register:
        The codec register to use for encoding the request. Defaults to
        :data:`~intessa.conneg.default.DEFAULT_REGISTER`.
    """

    headers = CaselessDictionary(headers)
    if not ((isinstance(data, str) and 'content-type' in headers) or
            ((hasattr(data, 'read') or hasattr(data, '__iter__')) and
             'content-type' in headers and
             'content-length' in headers)):
        if type is None:
            raise TypeError(
                "Request data provided with no content type information")
        headers['content-type'], data = codec_register.encode(type, data)
    return headers, data



"""Utilities for creating and manipulating HTTP ``Accept`` headers."""

from intessa.conneg import ContentType, DEFAULT_REGISTER


def make_accept_header(spec, codec_register=DEFAULT_REGISTER):

    """
    Build an accept header.

    Works on single strings:

        >>> make_accept_header('text/html')
        'text/html'

    Works on iterables of strings:

        >>> make_accept_header(('text/html', 'application/json'))
        'text/html,application/json'

    Works on iterables of type/quality pairs:

        >>> make_accept_header((('text/html', 0.7), ('application/json', 0.9)))
        'text/html;q=0.7,application/json;q=0.9'

    And mixed input:

        >>> make_accept_header((('text/html', 0.7), 'application/json'))
        'text/html;q=0.7,application/json'

    Also works on aliases defined on the codec register. Set up an alias:

        >>> from intessa.conneg.default import DEFAULT_REGISTER
        >>> reg = DEFAULT_REGISTER.copy()
        >>> reg.alias('app_json', 'application/json')

    And then use it on its own, or in lists:

        >>> make_accept_header('app_json', reg)
        'application/json'
        >>> make_accept_header(('app_json', 'text/html'), reg)
        'application/json,text/html'
        >>> make_accept_header((('app_json', 0.2), 'text/html'), reg)
        'application/json;q=0.2,text/html'
    """

    def resolve_media_type(mt):
        try:
            return codec_register.get_codec(mt)[0]
        except KeyError:
            return ContentType(mt).media_type

    if isinstance(spec, basestring):
        return resolve_media_type(spec)

    output = []
    for c_type in spec:
        if isinstance(c_type, basestring):
            output.append(resolve_media_type(c_type))
        else:
            media_type = resolve_media_type(c_type[0])
            quality = c_type[1]
            output.append('%s;q=%.1f' % (media_type, quality))
    return ','.join(output)



# -*- coding: utf-8 -*-

import re

from urecord import Record

from intessa.conneg import ContentType
from intessa.conneg.default import DEFAULT_REGISTER

class Response(Record('api',
                      # The essence of an HTTP response:
                      'status_code',
                      'headers',
                      'content',
                      'codec_register')):

    """
    An HTTP response, created by calling an :class:`intessa.api.API`.

    The following attributes all need to be set when constructing a response.

    .. py:attribute:: api

        The :class:`intessa.api.API` instance which created this Response.

    .. py:attribute:: status_code

        An integer representing the HTTP status code of the response.

    .. py:attribute:: headers

        A case-insensitive dictionary (or dict-like object) containing header
        names and values.

    .. py:attribute:: content

        A bytestring (i.e. ``str`` instance) containing the body of the
        response.

    You may optionally set the following:

    .. py:attribute:: codec_register

        An instance of :class:`intessa.conneg.codec_base.CodecRegister`, to be
        used for decoding this response. Defaults to
        :data:`~intessa.conneg.default.DEFAULT_REGISTER`.

    All the other attributes are dynamically created by the class itself.
    """

    def __new__(cls, api, status_code, headers, content,
                codec_register=DEFAULT_REGISTER):
        return super(cls, cls).__new__(cls, api=api,
                                       status_code=status_code,
                                       headers=headers, content=content,
                                       codec_register=codec_register)

    def __repr__(self):
        if 'content-type' in self.headers:
            mimetype = self.headers['content-type'].split(';')[0]
            return '<intessa.Response[%d, %s]>' % (self.status_code, mimetype)
        return '<intessa.Response[%d]>' % (self.status_code,)

    @property
    def type(self):

        """
        A :class:`intessa.conneg.ContentType` for this response.

        Use this to implement conditional handling of various types.

        :raises: ``AttributeError`` if no Content-Type header was given.

        Example::

            >>> resp = Response(None, 200,
            ...     {'content-type': 'text/html; charset=utf-8'},
            ...     '<html></html>')
            >>> resp.type
            ContentType('text/html; charset=utf-8')
            >>> resp.type.media_type
            'text/html'
            >>> resp.type.params['charset']
            'utf-8'
        """

        if 'content-type' not in self.headers:
            raise AttributeError("Response does not have a content type")
        return ContentType(self.headers['content-type'])

    @property
    def value(self):

        ur"""
        The content of this HTTP response, decoded into a Python object.

        This method uses the :attr:`codec_register` to decode responses. An
        object is retrieved simply by passing the content type and content
        bytestring into ``codec_register.decode()``.

        Note that the object will be decoded once per access; store a reference
        to the returned object if you wish to use it more than once.

            >>> json = Response(None, 200,
            ...     {'content-type': 'application/json'},
            ...     '{"a": 1}')
            >>> json.value
            {'a': 1}

            >>> text = Response(None, 200,
            ...     {'content-type': 'text/plain; charset=utf-8'},
            ...     'H\xc3\xa9llo W\xc3\xb6rld')
            >>> text.value
            u'H\xe9llo W\xf6rld'
        """

        return self.codec_register.decode(self.type, self.content)


class ResponseException(Exception):

    """
    Raiseable responses (for non-successful HTTP responses).

    On the one hand, these exceptions can be used in exactly the same way as
    :class:`~intessa.response.Response`:

        >>> response = Response(api=None, status_code=403,
        ...     headers={'content-type': 'application/json'},
        ...     content='{"error": "Forbidden"}')
        >>> exc = ResponseException(response)
        >>> exc
        <intessa.Error[403, application/json]>
        >>> exc.value
        {'error': 'Forbidden'}

    However, the main use of response exceptions is for raising and subsequent
    pattern matching using ``except`` statements:

        >>> try:
        ...     raise exc
        ... except ResponseException['4xx'], captured:
        ...     print "Captured!"
        Captured!

    In this case, for example, you can capture ``4xx``, ``40x`` and ``403``;
    the class hierarchy will be dynamically created and cached on
    :class:`ResponseException` itself:

        >>> isinstance(exc, ResponseException['4xx'])
        True
        >>> isinstance(exc, ResponseException['40x'])
        True
        >>> isinstance(exc, ResponseException['403'])
        True

    The class is aliased as ``intessa.Error`` for the sake of convenience:

        >>> import intessa
        >>> intessa.Error['403'] is intessa.ResponseException['403']
        True
    """

    exception_class_cache = {}

    class __metaclass__(type):

        def __getitem__(cls, status_code):
            status_code = str(status_code)
            if re.match(r'^[0-9]xx$', status_code):
                return cls.make_exception_class(status_code, parent=cls)
            elif re.match(r'^[0-9][0-9]x$', status_code):
                parent = cls.make_exception_class(status_code[:1] + 'xx', parent=cls)
                return cls.make_exception_class(status_code, parent=parent)
            elif re.match(r'^[0-9][0-9][0-9]$', status_code):
                parent1 = cls.make_exception_class(status_code[:1] + 'xx', parent=cls)
                parent2 = cls.make_exception_class(status_code[:2] + 'x', parent=parent1)
                return cls.make_exception_class(status_code, parent=parent2, final=True)
            raise KeyError("Not a valid status code pattern: %r" % (status_code,))

        def make_exception_hierarchy(cls, status):
            """Build the exception hierarchy for a given (full) status code."""

            status = str(status)
            if not re.match(r'^[0-9]{3}$', status):
                raise ValueError("Invalid status code format: %r" % (status,))

            matching_codes = [status[:1] + 'xx',
                              status[:2] + 'x',
                              status[:3]]
            return reduce(
                lambda parent, code: cls.make_exception_class(code, parent=parent),
                matching_codes, cls)

        def make_exception_class(cls, code, parent=None, final=False):
            """Create the exception class for a given status code pattern."""

            if code in cls.exception_class_cache:
                return cls.exception_class_cache[code]

            name = '%s[%s]' % (cls.__name__, code)

            if parent is None:
                parent = cls
            bases = (parent,)

            attrs = {}
            if final:
                attrs = {'__new__': Exception.__new__}

            exc_class = type(name, bases, attrs)
            cls.exception_class_cache[code] = exc_class
            return exc_class

    def __new__(cls, response):
        status = str(response.status_code)
        return cls[status](response)

    def __repr__(self):
        if 'content-type' in self.headers:
            mimetype = self.headers['content-type'].split(';')[0]
            return '<intessa.Error[%d, %s]>' % (self.status_code, mimetype)
        return '<intessa.Error[%d]>' % (self.status_code,)

    def __init__(self, response):
        self.response = response

    def __getattr__(self, attr):
        return getattr(self.response, attr)



from urecord import Record

from intessa.conneg.content_type import ContentType


class CodecRegister(dict):

    r"""
    A dictionary mapping media types to :class:`Codecs <Codec>`.

    Keys should be media types as bytestrings, and values should be
    :class:`Codec` subclasses (or at least match the decode/encode interface).

    As an example, start by defining a codec::

        >>> import simplejson
        >>> class JSONCodec(Codec):
        ...     def encode(media_type, obj, **params):
        ...         return (media_type, simplejson.dumps(obj, **params))
        ...     def decode(content_type, bytes):
        ...         return simplejson.loads(bytes)

    Create a register and add the codec to it::

        >>> reg = CodecRegister()
        >>> reg['application/json'] = JSONCodec
        >>> reg['application/json'] is JSONCodec
        True

    You can then encode and decode objects using the same method signatures as
    on individual codecs. The codec register will dispatch based on the media
    type provided::

        >>> reg.encode('application/json', {"a": 1})
        ('application/json', '{"a": 1}')
        >>> reg.encode('application/json', {"a": 1}, indent=True)
        ('application/json', '{\n "a": 1\n}')
        >>> reg.decode('application/json', '{"a": 1}')
        {'a': 1}

    Since full media types can be rather unwieldy, you can also register
    shorter aliases which will be resolved to their full counterparts first::

        >>> reg.alias('json', 'application/json')
        >>> reg.encode('json', {"a": 1})
        ('application/json', '{"a": 1}')

    Note that this only applies to encoding; for decoding, it's assumed that
    full content type headers will be provided. If an alias is used as the
    content type for a decode, a ``KeyError`` will be raised::

        >>> reg.decode('json', '{"a": 1}')  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        KeyError: 'json'
    """

    class _Alias(str):
        """Sentinel class to represent a short alias to a longer media type."""
        pass

    def alias(self, alias_name, media_type):
        self[alias_name] = self._Alias(media_type)

    def get_codec(self, media_type):

        """
        Resolve a media type or alias to a codec and media type.

            >>> reg = CodecRegister()
            >>> reg['application/json'] = 123
            >>> reg.alias('json', 'application/json')
            >>> reg.get_codec('json')
            ('application/json', 123)
            >>> reg.get_codec('application/json')
            ('application/json', 123)

        This method will even deal with multiple levels of indirection,
        considering the last name in the chain to be the full media type::

            >>> reg.alias('json2', 'json')
            >>> reg.get_codec('json2')
            ('application/json', 123)
            >>> reg.alias('json3', 'json2')
            >>> reg.get_codec('json3')
            ('application/json', 123)

        This method also takes full Content-Type headers; it will remove
        parameters before doing the lookup::

            >>> reg.get_codec('application/json; param=value')
            ('application/json', 123)
        """

        last_ref = ContentType(media_type).media_type
        codec = self[last_ref]
        while isinstance(codec, self._Alias):
            last_ref = codec
            codec = self[last_ref]
        return (last_ref, codec)

    def copy(self):
        """A ``copy()`` that will return :class:`CodecRegister` instances."""

        return type(self)(self.iteritems())

    def encode(self, media_type, obj, **params):
        media_type, codec = self.get_codec(media_type)
        encoded = codec.encode(media_type, obj, **params)
        if isinstance(encoded, str):
            return (media_type, encoded)
        return encoded

    def decode(self, content_type, bytes):
        if not isinstance(content_type, ContentType):
            content_type = ContentType(content_type)

        codec = self[content_type.media_type]
        if isinstance(codec, self._Alias):
            raise KeyError(content_type.media_type)

        return codec.decode(content_type, bytes)


class Codec(object):

    """
    A superclass for implementing codecs.

    This class should be considered abstract (:meth:`encode` and :meth:`decode`
    are unimplemented), but it has a metaclass which will make those methods
    static.

    Example::

        >>> import simplejson
        >>> class JSONCodec(Codec):
        ...     def encode(media_type, obj, **params):
        ...         return (media_type, simplejson.dumps(obj, **params))
        ...     def decode(content_type, bytes):
        ...         return simplejson.loads(bytes)
        >>> JSONCodec.encode('application/json', {"a": 1})
        ('application/json', '{"a": 1}')
        >>> JSONCodec.decode('application/json', '{"a": 1}')
        {'a': 1}

    Note that ``encode()`` and ``decode()`` do not take ``self``; the metaclass
    will make them static methods automatically. This allows you to register
    the classes directly on a :class:`CodecRegister`.
    """

    class __metaclass__(type):
        def __new__(mcls, name, bases, attrs):
            if 'decode' in attrs:
                attrs['decode'] = staticmethod(attrs['decode'])
            if 'encode' in attrs:
                attrs['encode'] = staticmethod(attrs['encode'])
            return super(mcls, mcls).__new__(mcls, name, bases, attrs)

    def encode(media_type, obj, **params):

        """
        Encode a Python object into the given media type.

        :param media_type:
            The desired ``type/subtype`` media type for the output.
        :param obj:
            The Python object to encode.
        :params: Any additional parameters for the encoder.
        :returns:
            A two-tuple of ``(content_type, bytes)``, where ``content_type``
            should be a ``str`` of the full ``Content-Type`` header (including
            parameters) and ``bytes`` a ``str`` of the encoded output.
        """

        raise NotImplementedError

    def decode(content_type, bytes):

        """
        Decode a bytestring to a Python object, given its content-type.

        :param intessa.conneg.ContentType content_type:
            The declared ``Content-Type`` header from the HTTP server.
        :param str bytes:
            A string of the content itself.
        :returns: A Python object.
        """

        raise NotImplementedError



# -*- coding: utf-8 -*-

"""Houses the default global codec register and codecs."""

import cgi
import urllib

import lxml.etree
import lxml.objectify
import simplejson
import webob.multidict

from intessa.conneg.codec_base import CodecRegister, Codec
from intessa.conneg.content_type import ContentType


DEFAULT_REGISTER = CodecRegister()


class TextCodec(Codec):

    """Default text codec."""

    def encode(media_type, string, encoding='utf-8', errors='strict'):

        ur"""
        Encode a unicode string as a bytestring using an encoding.

        :param encoding:
            The encoding to use (default: ``'utf-8'``).
        :param errors:
            The strategy for handling encoding errors (default: ``'strict'``).
            See the documentation on the built-in ``unicode.encode()`` for more
            information about this option.

            >>> TextCodec.encode('text/plain', u"Héllo Wörld")
            (ContentType('text/plain; charset=utf-8'), 'H\xc3\xa9llo W\xc3\xb6rld')
            >>> TextCodec.encode('text/plain', u"Héllo Wörld", encoding='latin1')
            (ContentType('text/plain; charset=latin1'), 'H\xe9llo W\xf6rld')
        """

        encoded = string.encode(encoding, errors)
        c_type = ContentType('%s; charset=%s' % (media_type, encoding))
        return (c_type, encoded)

    def decode(c_type, bytes):

        ur"""
        Decode a bytestring to unicode, using the content type's charset.

            >>> TextCodec.decode(ContentType('text/plain; charset=utf-8'),
            ...                  'H\xc3\xa9llo W\xc3\xb6rld')
            u'H\xe9llo W\xf6rld'
            >>> TextCodec.decode(ContentType('text/plain; charset=latin1'),
            ...                              'H\xe9llo W\xf6rld')
            u'H\xe9llo W\xf6rld'

        If no charset is present, this method assumes the input is UTF-8::

            >>> TextCodec.decode(ContentType('text/plain'),
            ...                  'H\xc3\xa9llo W\xc3\xb6rld')
            u'H\xe9llo W\xf6rld'

        The decoder always uses 'strict' error handling::

            >>> TextCodec.decode(ContentType('text/plain; charset=us-ascii'), # doctest: +ELLIPSIS
            ...                  'H\xc3\xa9llo W\xc3\xb6rld')
            Traceback (most recent call last):
            ...
            UnicodeDecodeError: 'ascii' codec can't decode byte 0xc3 in position 1: ordinal not in range(128)
        """

        return bytes.decode(c_type.params.get('charset', 'utf-8'))

DEFAULT_REGISTER['text/plain'] = TextCodec
DEFAULT_REGISTER.alias('text', 'text/plain')


class SimpleFormCodec(Codec):

    """Codec for simple application/x-www-form-urlencoded HTML forms."""

    def encode(media_type, obj, encoding='utf-8', errors='strict'):

        u"""
        Encode a simple dictionary of strings to an HTML form.

        :param encoding:
            The encoding to use (default: ``'utf-8'``).
        :param errors:
            The strategy for handling encoding errors (default: ``'strict'``).
            See the documentation on the built-in ``unicode.encode()`` for more
            information about this option.

        The simplest case, dictionaries of bytestrings, works:

            >>> SimpleFormCodec.encode('application/x-www-form-urlencoded', # doctest: +ELLIPSIS
            ...     {'a': 'b', 'c': 'd'})
            (ContentType(...), 'a=b&c=d')

        You can also pass in lists of key, value pairs:

            >>> SimpleFormCodec.encode('application/x-www-form-urlencoded', # doctest: +ELLIPSIS
            ...     [('a', 'b'), ('c', 'd')])
            (ContentType(...), 'a=b&c=d')

        Unicode strings will be encoded according to the ``encoding`` and
        ``errors`` parameters:

            >>> SimpleFormCodec.encode('application/x-www-form-urlencoded', # doctest: +ELLIPSIS
            ...     [(u'héllo', u'wørld')])
            (ContentType('...; charset=utf-8'), 'h%C3%A9llo=w%C3%B8rld')
        """

        c_type = ContentType('%s; charset=%s' % (media_type, encoding))

        if hasattr(obj, 'iteritems'):
            pairs = obj.iteritems()
        else:
            pairs = iter(obj)
        encoded_pairs = list(SimpleFormCodec.encode_pairs(pairs, encoding=encoding, errors=errors))
        data = urllib.urlencode(encoded_pairs)

        return (c_type, data)

    @staticmethod
    def encode_pairs(pairs, encoding='utf-8', errors='strict'):
        encoder = lambda s: s.encode(encoding, errors=errors)
        for key, value in pairs:
            if isinstance(key, unicode):
                key = encoder(key)
            if isinstance(value, unicode):
                value = encoder(value)
            yield (key, value)

    def decode(c_type, bytes):

        r"""
        Decode the encoded form, returning a ``MultiDict``.

        For content types without a ``charset``, returns an instance of
        ``webob.multidict.MultiDict`` with bytestrings:

            >>> form_type = ContentType('application/x-www-form-urlencoded')
            >>> SimpleFormCodec.decode(form_type, 'a=b&c=d')
            MultiDict([('a', 'b'), ('c', 'd')])

        If a charset is given, the input is decoded using it, returning a
        ``UnicodeMultiDict`` instead:

            >>> utf8_form_type = ContentType(
            ...     'application/x-www-form-urlencoded; charset=utf-8')
            >>> SimpleFormCodec.decode(utf8_form_type, 'h%C3%A9llo=w%C3%B8rld')
            UnicodeMultiDict([(u'h\xe9llo', u'w\xf8rld')])
        """

        pairs = cgi.parse_qsl(bytes)
        multi_dict = webob.multidict.MultiDict(pairs)
        if 'charset' in c_type.params:
            encoding = c_type.params['charset']
            multi_dict = webob.multidict.UnicodeMultiDict(multi_dict,
                                                          encoding=encoding,
                                                          decode_keys=True)
        return multi_dict

DEFAULT_REGISTER['application/x-www-form-urlencoded'] = SimpleFormCodec
DEFAULT_REGISTER.alias('form', 'application/x-www-form-urlencoded')


class JSONCodec(Codec):

    """Default JSON codec, powered by simplejson."""

    def encode(media_type, obj, **params):

        r"""
        Encode an object as JSON using ``simplejson.dumps()``.

        Additional parameters will be passed to ``simplejson.dumps()`` as-is.

            >>> JSONCodec.encode('application/json', {"a": 1})
            ('application/json', '{"a": 1}')
            >>> JSONCodec.encode('application/json', {"a": 1}, indent=True)
            ('application/json', '{\n "a": 1\n}')
        """

        return (media_type, simplejson.dumps(obj, **params))

    def decode(content_type, bytes):

        """
        Decode a JSON bytestring using ``simplejson.loads()``.

            >>> JSONCodec.decode('application/json', '{"a": 1}')
            {'a': 1}
        """

        return simplejson.loads(bytes)

DEFAULT_REGISTER['application/json'] = JSONCodec
DEFAULT_REGISTER['text/javascript'] = JSONCodec
DEFAULT_REGISTER.alias('json', 'application/json')
DEFAULT_REGISTER.alias('json-js', 'text/javascript')


class XMLCodec(Codec):

    """Default XML codec, using ``lxml.objectify``."""

    def encode(media_type, etree, **params):

        ur"""
        Encode an lxml ``ElementTree`` using ``lxml.etree.tostring()``.

        By default, the output will include an XML prolog, and will be
        utf-8-encoded. This can be overridden by passing ``xml_declaration``
        (``True`` or ``False``, default ``True``) and ``encoding`` (default
        ``'utf-8'``) keywords.

        See http://lxml.de/api/lxml.etree-module.html#tostring for a detailed
        overview of available options.

            >>> tree = lxml.etree.fromstring('<obj><attr>value</attr></obj>')
            >>> tree.find('attr').text = u"vålúè"
            >>> XMLCodec.encode('application/xml', tree)
            ('application/xml', "<?xml version='1.0' encoding='utf-8'?>\n<obj><attr>v\xc3\xa5l\xc3\xba\xc3\xa8</attr></obj>")
        """

        params.setdefault('xml_declaration', True)
        params.setdefault('with_tail', False)
        if params['xml_declaration']:
            params.setdefault('encoding', 'utf-8')

        return (media_type, lxml.etree.tostring(etree, **params))

    def decode(content_type, bytes):

        """
        Decode an XML bytestring to a Python object, using ``lxml.objectify``.

        For more information on these objects, see
        http://lxml.de/objectify.html.

            >>> doc = XMLCodec.decode('application/xml', '<obj><attr>value</attr></obj>')
            >>> doc # doctest: +ELLIPSIS
            <Element obj at 0x...>
            >>> doc.tag
            'obj'
            >>> doc.attr
            'value'
        """

        return lxml.objectify.fromstring(bytes)

DEFAULT_REGISTER['application/xml'] = XMLCodec
DEFAULT_REGISTER.alias('xml', 'application/xml')



"""This module contains representations of internet media type strings."""

import email.message
import mimetypes

import odict

# Load the default MIME types database.
if not mimetypes.inited:
    mimetypes.init()
if hasattr(mimetypes, '_db'):
    DEFAULT_MIMETYPES_DB = mimetypes._db
else:
    DEFAULT_MIMETYPES_DB = mimetypes.guess_extension.im_self

__all__ = ['ContentType']


class ContentType(str):

    """
    An HTTP ``Content-Type`` header.

    This is a subclass of ``str``, not ``unicode``, in accordance with RFC 2616
    (http://www.ietf.org/rfc/rfc2616), which states that only US-ASCII is to be
    used in Content-Type headers.

    Create a header from a bytestring::

        >>> c_type = ContentType('text/html; charset=utf-8')
        >>> c_type
        ContentType('text/html; charset=utf-8')

    Access the media type on its own::

        >>> c_type.media_type
        'text/html'

    And the parameters, as an ordered dictionary::

        >>> c_type.params
        odict([('charset', 'utf-8')])
    """

    def __repr__(self):
        return 'ContentType(%r)' % (self._normalize(),)

    def __str__(self):
        return self._normalize()

    def __eq__(self, other):

        """
        Determine equality, preserving order and ignoring whitespace.

            >>> ContentType('text/html') == 'text/html'
            True
            >>> ContentType('text/html; k=v') == 'text/html; k=v'
            True
            >>> ContentType('text/html; a=1; b=2') == 'text/html; b=2; a=1'
            False
            >>> ContentType('text/html  ; a=1 ; b=2') == 'text/html; a=1; b=2'
            True
        """

        self_parsed = self._parse(self)
        other_parsed = self._parse(str(other))
        return self_parsed == other_parsed

    @staticmethod
    def _parse(string):

        """
        Parse a Content-Type header using ``email.message.Message``.

        :returns: ``[('type/subtype', ''), ('param1', 'value1'), ...]``.
        """

        msg = email.message.Message()
        msg.add_header('Content-Type', string)
        return msg.get_params()

    def _normalize(self):

        """
        Normalize whitespace in this Content-Type header.

        Example::

            >>> ContentType('text/html;a=1;b=2')._normalize()
            'text/html; a=1; b=2'
            >>> ContentType('text/html  ;  a=1    ;  b=2')._normalize()
            'text/html; a=1; b=2'
        """

        params = '; '.join('='.join(pair) for pair in self.params.iteritems())
        return '%s; %s' % (self.media_type, params)

    @property
    def media_type(self):

        """
        Just the ``type/subtype`` pair of this content type.

            >>> ContentType('text/html; charset=utf-8').media_type
            'text/html'
        """

        return self._parse(self)[0][0]

    @property
    def params(self):

        """
        An ordered dict of the ``key=value`` parameters of this header.

            >>> ContentType('text/html; charset=utf-8').params
            odict([('charset', 'utf-8')])
            >>> ContentType('text/html; charset=utf-8').params['charset']
            'utf-8'
            >>> ContentType('text/html; charset=utf-8; boundary=xxx').params
            odict([('charset', 'utf-8'), ('boundary', 'xxx')])

        Note that each time you access this attribute a new dictionary will be
        created; modifications are unrecommended and will not persist.
        """

        return odict.odict(self._parse(self)[1:])

    def guess_extensions(self, db=DEFAULT_MIMETYPES_DB):

        """
        Return a list of likely file extensions for this media type.

        :keyword db: An instance of ``mimetypes.MimeTypes``; if unspecified,
                     will use the global default.
        :returns: A list of bytestrings, which will be empty if no matches were
                  found.

        Example::

            >>> ContentType('text/html').guess_extensions()
            ['.html', '.htm']
            >>> ContentType('application/x-will-not-match').guess_extensions()
            []
        """

        return db.guess_all_extensions(self.media_type) or []

    def guess_extension(self, db=DEFAULT_MIMETYPES_DB):
        """
        Return a likely file extension for this media type, or ``None``.

        :keyword db: An instance of ``mimetypes.MimeTypes``; if unspecified,
                     will use the global default.
        :returns: A single bytestring or ``None``.

        Example::

            >>> ContentType('text/html').guess_extension()
            '.html'
            >>> ContentType('application/x-will-not-match').guess_extension()
        """

        return db.guess_extension(self.media_type)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from distribute_setup import use_setuptools
use_setuptools()

from setuptools import setup, find_packages
import os.path as p

VERSION = open(p.join(p.dirname(p.abspath(__file__)), 'VERSION')).read().strip()

setup(
    name='intessa',
    version=VERSION,
    description='Zero-configuration dynamic interfaces to HTTP APIs.',
    author='Zachary Voase',
    author_email='z@zacharyvoase.com',
    url='http://github.com/zacharyvoase/intessa',
    packages=find_packages(where='lib'),
    package_dir={'': 'lib'},
    install_requires=[
        'urlobject==0.6.0',
        'urecord==0.0.3',
        'requests==0.8.0',
        'odict==1.4.1',
        'simplejson==2.1.6',
        'lxml==2.3',
        'webob',
        'mock==0.7.2',
        'nose==1.1.2',
    ],
)



"""
Code responsible for low-level HTTP interactions.

At the moment this just contains the default HTTP backend for
:class:`intessa.API`, based on the requests_ library for Python.

.. _requests: http://python-requests.org/
"""

from requests import request, exceptions

from intessa.conneg.default import DEFAULT_REGISTER
from intessa.response import Response, ResponseException


def requests_http_backend(api, method, url, *args, **kwargs):
    """Shim over ``requests.request`` to return :class:`Responses <intessa.response.Response>`."""

    codec_register = kwargs.pop('codec_register', DEFAULT_REGISTER)
    http_response = request(method, url, *args, **kwargs)

    if not http_response.status_code and http_response.error:
        if isinstance(http_response.error, exceptions.Timeout):
            raise http_response.error.args[0]
        raise http_response.error

    response = Response(api=api,
                        status_code=http_response.status_code,
                        headers=http_response.headers,
                        content=http_response.content,
                        codec_register=codec_register)

    if (response.status_code // 100) != 2:
        raise ResponseException(response)
    return response
